/*
mpcx_tree

high-level functions working with trees of polynomials

Copyright (C) 2009, 2010 Andreas Enge

This file is part of the MPFRCX Library.

The MPFRCX Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The MPFRCX Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the MPC Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#include "mpfrcx-impl.h"

static mpcx_t** mpcx_subproducttree (mpcx_t *factors, int no);
static mpcx_t** mpcx_evaltree (mpcx_t **subproducts, int no, mpcx_srcptr f);
static void mpcx_multieval_fast (mpc_ptr *values, mpc_srcptr *args, int no,
                                 mpcx_srcptr f);

/**************************************************************************/

void mpcx_reconstruct (mpcx_ptr rop, mpcx_t *factors, int no, int verbose)
   /* takes an array of no polynomials and multiplies them all together   */
   /* into rop                                                            */

{
   int i, layer = 1;
   mpcx_t *op;
   timer clock;

   op = (mpcx_t *) malloc (no * sizeof (mpcx_t));
   /* copy factors into op */
   for (i = 0; i < no; i++)
      mpcx_init_set (op [i], factors [i]);

   while (no > 1) {
      timer_start (clock);
      for (i = 0; i < no / 2; i++)
         /* replace op [i] by op [2i] * op [2i+1] */
         mpcx_mul (op [i], op [2*i], op [2*i+1]);
      if (no % 2 != 0)
         mpcx_swap (op [no / 2], op [no - 1]);
      for (i = (no + 1) / 2; i < no; i++)
         mpcx_clear (op [i]);
      no = (no + 1) / 2;
      timer_stop (clock);
      if (verbose)
         printf ("- Layer %2d: %.1f\n", layer, timer_get (clock));
      layer++;
   }

   /* The result is now in op [0]. */
   mpcx_set (rop, op [0]);
   mpcx_clear (op [0]);
   free (op);
}

/**************************************************************************/

static mpcx_t** mpcx_subproducttree (mpcx_t *factors, int no) {
   /* takes an array of "no" polynomials and returns the binary tree      */
   /* representing their subproducts: res [0,j] contains factors [j], and */
   /* res [i+1,j] contains res [i,2j] * res [i,2j+1]. If no is not a      */
   /* power of 2, then neither are the numbers of polynomials in each     */
   /* level.                                                              */
   /* Optimal (at least if all degrees are the same) version as found out */
   /* with Guillaume and Pierrick: always multiply two factors, sweeping  */
   /* the array from left to right                                        */
   /* Makes only sense if all polynomials have coefficients with the same */
   /* precision.                                                          */

   mpcx_t** res;
   int      i, j, layers, tmp, no_new, no_alt;

   for (tmp = no, layers = 1; tmp > 1; tmp = (tmp + 1) / 2, layers++);
   res = (mpcx_t **) malloc (layers * sizeof (mpcx_t *));

   /* construct the basic layer */
   no_new = no;
   res [0] = (mpcx_t *) malloc (no_new * sizeof (mpcx_t));
   for (j = 0; j < no_new; j++)
      mpcx_init_set (res [0][j], factors [j]);
   no_alt = no;

   for (i = 0; i < layers - 1; i++) {
      /* construct layer i+1 */
      no_new = (no_alt + 1) / 2;
      res [i+1] = (mpcx_t *) malloc (no_new * sizeof (mpcx_t));
      for (j = 0; j < no_new; j++)
         if (no_alt % 2 == 1 && j == no_new - 1)
            mpcx_init_set (res [i+1][j], res [i][2*j]);
         else {
            mpcx_init (res [i+1][j],
               res [i][2*j]->deg + res [i][2*j+1]->deg + 1,
               res [i][2*j]->prec);
            mpcx_mul (res [i+1][j], res [i][2*j], res [i][2*j+1]);
         }
      no_alt = no_new;
      fflush (stdout);
   }

   return res;
}
/**************************************************************************/

static mpcx_t** mpcx_evaltree (mpcx_t **subproducts, int no,
                               mpcx_srcptr f) {
   /* second step in multipoint evaluation; takes a subproduct tree with  */
   /* "no" entries and a polynomial f, and returns the tree of the        */
   /* f mod subproducts [i][j]                                            */
   /* Makes only sense if all polynomials have coefficients with the same */
   /* precision.                                                          */

   mpcx_t** res;
   int      i, j, layers, tmp;
   int*     nos;
      /* the number of entries in each layer */

   for (tmp = no, layers = 1; tmp > 1; tmp = (tmp + 1) / 2, layers++);
   nos = (int *) malloc (layers * sizeof (int));
   nos [0] = no;
   for (i = 0; i < layers - 1; i++)
      nos [i+1] = (nos [i] + 1) / 2;
   res = (mpcx_t **) malloc (layers * sizeof (mpcx_t *));
   for (i = 0; i < layers; i++) {
      res [i] = (mpcx_t *) malloc (nos [i] * sizeof (mpcx_t));
      for (j = 0; j < nos [i]; j++)
         mpcx_init (res [i][j], subproducts [i][j]->deg,
                     subproducts [i][j]->prec);
   }

   mpcx_rem (res [layers-1][0], f, subproducts [layers-1][0]);
   for (i = layers - 2; i >= 0; i--)
      for (j = 0; j < nos [i]; j++)
         mpcx_rem (res [i][j], res [i+1][j/2], subproducts [i][j]);

   free (nos);

   return res;
}

/**************************************************************************/

void mpcx_multieval (mpc_ptr *values, mpc_srcptr *args, int no, mpcx_srcptr f) {
   /* evaluates the polynomial f in the "no" arguments args and returns   */
   /* the results via values. values must contain sufficient space, and   */
   /* each entry must be initialised.                                     */
   /* values and args may point to the same array.                        */
   /* Works only if f and values have the same precision.                 */
   /* Splits the arguments into portions of at least half the degree of   */
   /* f and calls the fast multipoint evaluation routine on the chunks.   */

   int chunks, size;
   int i;

   printf ("\n--- multieval deg %i in %i arguments\n", f->deg, no);
   fflush (stdout);
   chunks = (2 * no) / f->deg;
   if (chunks == 0)
      chunks = 1;
   size = no / chunks;
   printf ("chunks %i\n", chunks);
   printf ("size %i\n", size);

   for (i = 0; i < chunks - 1; i++)
      mpcx_multieval_fast (values + i*size, args + i*size, size, f);
   mpcx_multieval_fast (values + (chunks - 1)*size, args + (chunks - 1)*size,
         no - (chunks - 1) * size, f);
}

/***************************************************************************/

static void mpcx_multieval_fast (mpc_ptr *values, mpc_srcptr *args, int no,
                                 mpcx_srcptr f) {
   /* The internal version of the function. Always builds a subproduct    */
   /* tree independently of the degree of f and the number of arguments.  */

   mpcx_t** subproducts;
   mpcx_t** evaltree;
   mpcx_t*  factors;
   int      i, j, layers, tmp;
   int*     nos;
      /* the number of entries in each layer */

   printf ("\n--- multieval_fast deg %i in %i arguments\n", f->deg, no);
   fflush (stdout);
   for (tmp = no, layers = 1; tmp > 1; tmp = (tmp + 1) / 2, layers++);
   nos = (int *) malloc (layers * sizeof (int));
   nos [0] = no;
   for (i = 0; i < layers - 1; i++)
      nos [i+1] = (nos [i] + 1) / 2;

   factors = (mpcx_t *) malloc (no * sizeof (mpcx_t));
   for (i = 0; i < no; i++) {
      mpcx_init (factors [i], 2, f->prec);
      factors [i]->deg = 1;
      mpc_set_ui (factors [i]->coeff [1], 1, MPC_RNDNN);
      mpc_neg (factors [i]->coeff [0], args [i], MPC_RNDNN);
   }

   subproducts = mpcx_subproducttree (factors, no);
   fflush (stdout);

   evaltree = mpcx_evaltree (subproducts, no, f);
   for (j = 0; j < no; j++)
      mpc_set (values [j], evaltree [0][j]->coeff [0], MPC_RNDNN);

   for (i = 0; i < layers; i++) {
      for (j = 0; j < nos [i]; j++) {
         mpcx_clear (subproducts [i][j]);
         mpcx_clear (evaltree [i][j]);
      }
      free (subproducts [i]);
      free (evaltree [i]);
   }
   free (subproducts);
   free (evaltree);
   for (i = 0; i < no; i++)
      mpcx_clear (factors [i]);
   free (factors);
   free (nos);
}
